# coding:utf-8
'''
Test for functional.im2col
Created   :   7, 26, 2018
Revised   :   7, 26, 2018
All rights reserved
'''
__author__ = 'dawei.leng'

import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.functional import _im2col as im2col
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)


def _test_case_0():
    print('test case 0')
    import numpy as np
    x = tensor.ftensor4()
    y = im2col(x, nb_size=(1,1), border_mode='valid', merge_channel=False)
    f = theano.function([x], y, no_default_updates=True)

    # B, C, H, W = 1, 1, 3, 3
    for i in range(10):
        B, C, H, W = np.random.randint(1, 8), np.random.randint(1, 4), np.random.randint(3, 128), np.random.randint(3, 128)
        X = np.random.rand(B, C, H, W).astype('float32')
        Y = f(X)
        print('X.shape=', X.shape)
        print('Y.shape=', Y.shape)
        if np.all(X==Y[:,:,:,:,0]):
            pass
        else:
            diff = np.max(abs(X-Y[:,:,:,:,0]))
            print('max diff=', diff)
            print('X=', X.flatten())
            print('Y=', Y[:,:,:,:,0].flatten())
            raise ValueError('X!=Y')

def _test_case_1():
    print('test case 1')
    import numpy as np
    x = tensor.ftensor4()
    y = im2col(x, nb_size=(1,1), border_mode='valid', merge_channel=True)
    f = theano.function([x], y, no_default_updates=True)

    # B, C, H, W = 1, 1, 3, 3
    for i in range(10):
        B, C, H, W = np.random.randint(1, 8), np.random.randint(1, 4), np.random.randint(3, 128), np.random.randint(3, 128)
        X = np.random.rand(B, C, H, W).astype('float32')
        Y = f(X)
        Y2 = X.transpose((0, 2, 3, 1))
        # Y2 = np.reshape(Y2, (B, H, W, -1))
        print('Y.shape=', Y.shape)
        print('Y2.shape=', Y2.shape)
        if np.all(Y==Y2):
            pass
        else:
            diff = np.max(abs(Y-Y2))
            print('max diff=', diff)
            print('Y=',  Y.flatten())
            print('Y2=', Y2.flatten())
            raise ValueError('X!=Y')

def _test_case_2():
    print('test case 2')
    import numpy as np
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 6, 6, (3,3), (1,1), 'valid' # failed
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 6, 6, (2,2), (1,1), 'valid' # failed
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 5, 6, (3,3), (1,1), 'half' # pass, mode half need neighbour with odd shapes
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 6, 6, (2,2), (1,1), 'half' # failed
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 9, 9, (3,3), (1,1), 'ignore_borders' # failed
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 6, 6, (3,3), (1,1), 'full' # failed
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 5, 6, (2,2), (1,1), 'wrap_centered' # failed, mode wrap_centered need neighbour with odd shapes
    # B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 5, 6, (3,3), (1,1), 'wrap_centered' # pass, mode wrap_centered need neighbour with odd shapes
    B, C, H, W, nb_size, nb_step, border_mode = 2, 4, 10, 9, (3,3), (1,1), 'wrap_centered' # pass, mode wrap_centered need neighbour with odd shapes

    x = tensor.ftensor4()
    y = im2col(x, nb_size=nb_size, border_mode=border_mode, merge_channel=False)
    f = theano.function([x], y, no_default_updates=True)

    X = np.random.rand(B, C, H, W).astype('float32')
    Y = f(X)
    print('X.shape=', X.shape)
    print('Y.shape=', Y.shape)
    if X.shape[0] == Y.shape[0] and X.shape[1] == Y.shape[1] and X.shape[2] == Y.shape[2] and  X.shape[3] == Y.shape[3] and Y.shape[-1] == nb_size[0] * nb_size[1]:
        pass
    else:
        raise ValueError('Shape not consistent')



if __name__ == '__main__':

    _test_case_0()
    _test_case_1()
    _test_case_2()

    print('Test passed')






# coding:utf-8
'''
Test for CTPN implementation
Created   :   7, 27, 2018
Revised   :   7, 27, 2018
All rights reserved
'''
__author__ = 'dawei.leng'

import os
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"
# os.environ['THEANO_FLAGS'] = "floatX=float32, mode=DEBUG_MODE, warn_float64='raise', exception_verbosity=high"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.model.ctpn import model_CTPN
from dandelion.objective import *

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    model = model_CTPN(k=10, do_side_refinement_regress=False,
                 batchnorm_mode=1, channel=3, im_height=None, im_width=None,
                 kernel_size=3, border_mode=(1, 1), VGG_flip_filters=False,
                 im2col=None)
    x = tensor.ftensor4('x')
    y1 = tensor.ftensor5('y1')
    y2 = tensor.ftensor4('y2')
    class_score, bboxs = model.forward(x)
    #--- check back-prop ---#
    loss = aggregate(squared_error(y1, class_score)) + aggregate(squared_error(y2, bboxs))
    grad = theano.grad(loss, model.collect_params())
    print('back-prop test pass')


    print('compiling fn...')
    fn = theano.function([x], [class_score, bboxs], no_default_updates=False, on_unused_input='ignore')
    print('run fn...')
    input = np.random.rand(4, 3, 256, 256).astype(np.float32)
    class_score, bboxs = fn(input)
    assert class_score.shape == (4, 16, 16, 10, 2), 'class_score shape not correct'
    assert bboxs.shape == (4, 16, 16, 10, 2), 'bboxs shape not correct'

    # print(class_score.shape)
    # print(bboxs.shape)



if __name__ == '__main__':

    test_case_0()

    print('Test passed')






# coding:utf-8
# Unit test for GroupNorm module
# Created   :  11, 22, 2018
# Revised   :  11, 22, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import *
from dandelion.objective import *
from dandelion.update import *
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def fix_update_bcasts(updates):
    for param, update in updates.items():
        if param.broadcastable != update.broadcastable:
            updates[param] = tensor.patternbroadcast(update, param.broadcastable)
    return updates

def test_case_0():
    B, C, H, W = 4, 128, 256, 256
    x = tensor.ftensor4('x')
    z = tensor.ftensor4('gt')
    conv0 = Conv2D(in_channels=2*C, out_channels=C, kernel_size=(3,3), pad='same')
    # gn0   = GroupNorm(channel_num=C, group_num=32, beta=None, gamma=None)
    gn0   = GroupNorm(channel_num=C, group_num=32)
    model = Sequential([conv0, gn0], activation=relu)
    y = model.forward(x)
    loss = aggregate(squared_error(y, z))
    updates = adadelta(loss, model.collect_params())
    updates.update(model.collect_self_updates())
    # f = theano.function([x], y, no_default_updates=False, updates=fix_update_bcasts(bn.collect_self_updates()))
    f = theano.function([x, z], [y, loss], no_default_updates=False, updates=updates)
    x = np.random.rand(B, 2*C, H, W).astype('float32')
    z = np.random.rand(B, C, H, W).astype('float32')
    y, loss = f(x, z)
    assert y.shape ==(B, C, H, W)
    print('test_case_0 passed')

if __name__ == '__main__':

    # test_case_0()

    test_case_0()

    print('Test passed')


# coding:utf-8
# Test for spatial_pyramid pooling
# Created   :   7,  5, 2018
# Revised   :   7,  5, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
# os.environ['THEANO_FLAGS'] = "floatX=float64, mode=FAST_RUN"
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.functional import spatial_pyramid_pooling
from lasagne.layers import InputLayer, get_output, SpatialPyramidPoolingLayer
import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

class build_model_D(Module):
    def __init__(self, pyramid_dims=[6, 4, 2, 1]):
        super().__init__()
        self.pyramid_dims = pyramid_dims
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, C, H, W)
        :return:
        """
        x = spatial_pyramid_pooling(x, pyramid_dims=self.pyramid_dims, implementation='fast_ls')
        return x

def build_model_L(pyramid_dims=[6, 4, 2, 1]):
    input_var = tensor.ftensor4('x')  # (B, C, H, W)
    input0 = InputLayer(shape=(None, None, None, None), input_var=input_var, name='input0')
    x = SpatialPyramidPoolingLayer(input0, pool_dims=pyramid_dims, implementation='fast')
    return x

def test_case_0():
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name
    np.random.seed(0)

    pyramid_dims = [3, 2, 1]

    model_D = build_model_D(pyramid_dims=pyramid_dims)
    model_L = build_model_L(pyramid_dims=pyramid_dims)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    for i in range(20):
        B = np.random.randint(low=1, high=33)
        C = np.random.randint(low=1, high=32)
        H = np.random.randint(low=5, high=512)
        W = np.random.randint(low=5, high=513)
        x = np.random.rand(B, C, H, W).astype(np.float32) - 0.5
        y_D= fn_D(x)
        y_L = fn_L(x)
        # print(y_D)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print(y_D.shape)
            print(y_L.shape)
            print('B=%d, C=%d, H=%d, W=%d\n' % (B, C, H, W))
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')


class build_model_D_1(Module):
    def __init__(self, pyramid_dims=[6, 4, 2, 1]):
        super().__init__()
        self.pyramid_dims = pyramid_dims
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, C, H, W)
        :return:
        """
        x = spatial_pyramid_pooling(x, pyramid_dims=self.pyramid_dims, implementation='fast')
        return x

def build_model_L_1(pyramid_dims=[6, 4, 2, 1]):
    input_var = tensor.ftensor4('x')  # (B, C, H, W)
    input0 = InputLayer(shape=(None, None, None, None), input_var=input_var, name='input0')
    x = SpatialPyramidPoolingLayer(input0, pool_dims=pyramid_dims, implementation='fast')
    return x


def test_case_1():
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name
    np.random.seed(0)

    pyramid_dims = [4]

    model_D = build_model_D_1(pyramid_dims=pyramid_dims)
    model_L = build_model_L_1(pyramid_dims=pyramid_dims)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    # y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    # fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    x = np.array([[[[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]]])
    x = x.astype(np.float32)
    # y = fn_L(x)  #  Lasagne will crash
    y = fn_D(x)
    y_expect = np.array([[[[0.1, 0.2, 0.3, 0.0], [0.4, 0.5, 0.6, 0.0], [0.7, 0.8, 0.9, 0.0], [0.0, 0.0, 0.0, 0.0]]]])
    y_expect = y_expect.astype(np.float32)
    y_expect = y_expect.reshape((1, 1, -1))
    diff = np.max(np.abs(y - y_expect))
    print('diff=%0.6f' % diff)
    if diff > 1e-4:
        raise ValueError('diff is too big')


class build_model_D_2(Module):
    def __init__(self, pyramid_dims=[6, 4, 2, 1]):
        super().__init__()
        self.pyramid_dims = pyramid_dims
        self.predict = self.forward

    def forward(self, x):
        """

        :param x: (B, C, H, W)
        :return:
        """
        x = spatial_pyramid_pooling(x, pyramid_dims=self.pyramid_dims, implementation='stretch')
        return x

def build_model_L_2(pyramid_dims=[6, 4, 2, 1]):
    input_var = tensor.ftensor4('x')  # (B, C, H, W)
    input0 = InputLayer(shape=(None, None, None, None), input_var=input_var, name='input0')
    x = SpatialPyramidPoolingLayer(input0, pool_dims=pyramid_dims, implementation='kaiming')
    return x


def test_case_2():
    import numpy as np
    from lasagne_ext.utils import get_layer_by_name
    np.random.seed(0)

    pyramid_dims = [4]

    model_D = build_model_D_2(pyramid_dims=pyramid_dims)
    model_L = build_model_L_2(pyramid_dims=pyramid_dims)

    X = get_layer_by_name(model_L, 'input0').input_var
    y_D = model_D.forward(X)
    # y_L = get_output(model_L)

    fn_D = theano.function([X], y_D, no_default_updates=True, on_unused_input='ignore')
    # fn_L = theano.function([X], y_L, no_default_updates=True, on_unused_input='ignore')

    x = np.array([[[[0.1, 0.2], [0.3, 0.4]]]])
    x = x.astype(np.float32)
    # y = fn_L(x)  #  Lasagne will use float64
    y = fn_D(x)
    y_expect = np.array([[[[0.1, 0.1, 0.2, 0.2], [0.1, 0.1, 0.2, 0.2], [0.3, 0.3, 0.4, 0.4], [0.3, 0.3, 0.4, 0.4]]]])
    y_expect = y_expect.astype(np.float32)
    y_expect = y_expect.reshape((1, 1, -1))
    diff = np.max(np.abs(y - y_expect))
    print('diff=%0.6f' % diff)
    if diff > 1e-4:
        raise ValueError('diff is too big')

    x = np.array([[[[0.7, 0.2, 0.3], [0.5, 0.6, 0.4], [0.1, 0.9, 0.8]]]])
    x = x.astype(np.float32)
    y = fn_D(x)
    y_expect = np.array([[[[0.7, 0.7, 0.3, 0.3], [0.7, 0.7, 0.6, 0.4], [0.5, 0.9, 0.9, 0.8], [0.1, 0.9, 0.9, 0.8]]]])
    y_expect = y_expect.astype(np.float32)
    y_expect = y_expect.reshape((1, 1, -1))
    diff = np.max(np.abs(y - y_expect))
    print('diff=%0.6f' % diff)
    if diff > 1e-4:
        raise ValueError('diff is too big')


if __name__ == '__main__':

    test_case_0()

    test_case_1()

    test_case_2()

    print('Test passed')






# coding:utf-8
# Unit test for pooling functions
# Created   :   2, 27, 2018
# Revised   :   2, 27, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, sys
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.functional import *

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def pool_1d_Lasagne(x, axis=1, mode='max'):
    """
    Lasagne requires x is 3D, and pooling is done on the last dimension
    :param x:
    :param axis:
    :return:
    """
    input_4d = tensor.shape_padright(x, 1)
    if axis == 1:
        input_4d = input_4d.dimshuffle((0, 2, 1, 3))
    pooled = pool_2d(input_4d,
                     ws=(2, 1),
                     stride=(2, 1),
                     ignore_border=True,
                     pad=(0, 0),
                     mode=mode,
                     )
    if axis == 1:  # [DV] add support for 'axis' para
        pooled = pooled.dimshuffle((0, 2, 1, 3))
    return pooled[:, :, :, 0]

def test_case_0():
    import numpy as np

    x_3d = tensor.ftensor3('x')
    y_3d_D = pool_1d(x_3d, axis=1)

    y_3d_L = pool_1d_Lasagne(x_3d, axis=1)

    fn_D = theano.function([x_3d], y_3d_D, no_default_updates=True, on_unused_input='ignore')
    fn_L = theano.function([x_3d], y_3d_L, no_default_updates=True, on_unused_input='ignore')


    for i in range(20):
        x = np.random.rand(7, 117, 27).astype(np.float32)
        y_D = fn_D(x)
        y_L = fn_L(x)
        diff = np.max(np.abs(y_D - y_L))
        print('i=%d, diff=%0.6f' % (i, diff))
        if diff>1e-4:
            print('y_D=\n', y_D)
            print('y_L=\n', y_L)
            raise ValueError('diff is too big')

if __name__ == '__main__':

    test_case_0()

    print('Test passed')






# coding:utf-8
'''
Model definition of [feature pyramid network](https://arxiv.org/abs/1612.03144)
Created   :   7,  6, 2018
Revised   :   7,  6, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import theano.tensor as tensor
from ..module import *
from ..functional import *
from ..activation import *
from .resnet import ResNet_bottleneck


class model_FPN(Module):
    """
    Reference implementation of [feature pyramid network](https://arxiv.org/abs/1612.03144)
    return tuple of (p2, p3, p4, p5), cnn pyramid features at different scales
    """

    def __init__(self, input_channel=3, base_n_filters=64, batchnorm_mode=1):
        """
        8 effective conv in depth
        :param input_channel:
        :param base_n_filters:
        :param batchnorm_mode: passed to `ResNet_bottleneck()`
        """
        super().__init__()
        self.conv1 = Conv2D(in_channels=input_channel, out_channels=base_n_filters, kernel_size=3, pad='same', stride=(2, 2))
        self.bn1   = BatchNorm(input_shape=(None, base_n_filters, None, None))
        self.conv2 = Conv2D(in_channels=base_n_filters, out_channels=base_n_filters * 2, kernel_size=3, pad='same', stride=(2, 2))
        self.bn2   = BatchNorm(input_shape=(None, base_n_filters * 2, None, None))
        self.conv3 = Conv2D(in_channels=base_n_filters * 2, out_channels=base_n_filters * 4, kernel_size=3, pad='same', stride=(2, 2))
        self.bn3   = BatchNorm(input_shape=(None, base_n_filters * 4, None, None))
        self.conv4 = Conv2D(in_channels=base_n_filters * 4, out_channels=base_n_filters * 8, kernel_size=3, pad='same', stride=(2, 2))
        self.bn4   = BatchNorm(input_shape=(None, base_n_filters * 8, None, None))

        self.res_block1 = ResNet_bottleneck(outer_channel=base_n_filters, inner_channel=base_n_filters // 2, batchnorm_mode=batchnorm_mode)
        self.res_block2 = ResNet_bottleneck(outer_channel=base_n_filters * 2, inner_channel=base_n_filters, batchnorm_mode=batchnorm_mode)
        self.res_block3 = ResNet_bottleneck(outer_channel=base_n_filters * 4, inner_channel=base_n_filters * 2, batchnorm_mode=batchnorm_mode)
        self.res_block4 = ResNet_bottleneck(outer_channel=base_n_filters * 8, inner_channel=base_n_filters * 4, batchnorm_mode=batchnorm_mode)

        self.conv5 = Conv2D(in_channels=base_n_filters * 8, out_channels=base_n_filters * 4, kernel_size=1, pad='same')
        self.conv6 = Conv2D(in_channels=base_n_filters * 4, out_channels=base_n_filters * 4, kernel_size=1, pad='same')
        self.conv7 = Conv2D(in_channels=base_n_filters * 2, out_channels=base_n_filters * 4, kernel_size=1, pad='same')
        self.conv8 = Conv2D(in_channels=base_n_filters,     out_channels=base_n_filters * 4, kernel_size=1, pad='same')

        self.conv9  = Conv2D(in_channels=base_n_filters * 4, out_channels=base_n_filters * 4, kernel_size=1, pad='same')
        self.conv10 = Conv2D(in_channels=base_n_filters * 4, out_channels=base_n_filters * 4, kernel_size=1, pad='same')
        self.conv11 = Conv2D(in_channels=base_n_filters * 4, out_channels=base_n_filters * 4, kernel_size=1, pad='same')

    def forward(self, x):
        """
        :param x: (B, C, H, W)
        :return:
        """
        self.work_mode = 'train'

        c1 = relu(self.bn1.forward(self.conv1.forward(x)))  # stride = 2
        c2 = relu(self.res_block1.forward(c1))

        c3 = relu(self.bn2.forward(self.conv2.forward(c2)))  # stride = 4
        c3 = relu(self.res_block2.forward(c3))

        c4 = relu(self.bn3.forward(self.conv3.forward(c3)))  # stride = 8
        c4 = relu(self.res_block3.forward(c4))

        c5 = relu(self.bn4.forward(self.conv4.forward(c4)))  # stride = 16
        c5 = relu(self.res_block4.forward(c5))

        p5 = relu(self.conv5.forward(c5))
        p4 = relu(self.conv6.forward(c4)) + upsample_2d(p5, ratio=2, mode='repeat')
        p3 = relu(self.conv7.forward(c3)) + upsample_2d(p4, ratio=2, mode='repeat')
        p2 = relu(self.conv8.forward(c2)) + upsample_2d(p3, ratio=2, mode='repeat')

        p4 = relu(self.conv9.forward(p4))
        p3 = relu(self.conv10.forward(p3))
        p2 = relu(self.conv11.forward(p2))

        return p2, p3, p4, p5

    def predict(self, x):
        self.work_mode = 'inference'

        c1 = relu(self.bn1.predict(self.conv1.predict(x)))  # stride = 2
        c2 = relu(self.res_block1.predict(c1))

        c3 = relu(self.bn2.predict(self.conv2.predict(c2)))  # stride = 4
        c3 = relu(self.res_block2.predict(c3))

        c4 = relu(self.bn3.predict(self.conv3.predict(c3)))  # stride = 8
        c4 = relu(self.res_block3.predict(c4))

        c5 = relu(self.bn4.predict(self.conv4.predict(c4)))  # stride = 16
        c5 = relu(self.res_block4.predict(c5))

        p5 = relu(self.conv5.predict(c5))
        p4 = relu(self.conv6.predict(c4)) + upsample_2d(p5, ratio=2, mode='repeat')
        p3 = relu(self.conv7.predict(c3)) + upsample_2d(p4, ratio=2, mode='repeat')
        p2 = relu(self.conv8.predict(c2)) + upsample_2d(p3, ratio=2, mode='repeat')

        p4 = relu(self.conv9.predict(p4))
        p3 = relu(self.conv10.predict(p3))
        p2 = relu(self.conv11.predict(p2))

        return p2, p3, p4, p5



# coding:utf-8
'''
  Model summary and visualization toolkits
  Created   :  11, 15, 2018
  Revised   :  11, 16, 2018  add `size_unit` arg for `get_model_summary()`
  All rights reserved
'''
__author__ = 'dawei.leng'

from collections import OrderedDict

def get_model_size(model):
    """
    Calculate model parameter size
    :param model:
    :return: total size of model, in bytes
    """
    weights = model.get_weights()
    model_size = 0
    for weight, _ in weights:
        model_size += weight.size * weight.itemsize
    return model_size

def get_model_summary(model, size_unit='M'):
    """
    Produce model parameter summary
    :param model:
    :param size_unit: {'M'|'K'|'B'|int}
    :return: OrderedDict
    """
    if isinstance(size_unit, str):
        size_unit = size_unit.upper()
        assert size_unit in {'M', 'K', 'B'}
        if size_unit == 'M':
            size_unit = 1048576
        elif size_unit == 'K':
            size_unit = 1024
        else:
            size_unit = 1
    model_summary = OrderedDict()
    model_summary['name'] = model.name
    model_size = get_model_size(model) / size_unit
    model_summary['size'] = model_size
    if len(model.params) > 0:
        model_summary['params'] = []
        for param in model.params:
            v = param.get_value()
            model_summary['params'].append(OrderedDict([('name',param.name), ('shape',str(v.shape)), ('dtype',v.dtype.name), ('size',v.size * v.itemsize/size_unit), ('percent',v.size * v.itemsize/(model_size*size_unit)*100)]))
    if len(model.self_updating_variables) > 0:
        model_summary['self_updating_variables'] = []
        for param in model.self_updating_variables:
            v = param.get_value()
            model_summary['self_updating_variables'].append(OrderedDict([('name',param.name), ('shape',str(v.shape)), ('dtype',v.dtype.name), ('size',v.size * v.itemsize/size_unit), ('percent',v.size * v.itemsize/(model_size*size_unit)*100)]))
    if len(model.sub_modules) > 0:
        sub_module_summaries = []
        for tag, child in (model.sub_modules.items()):
            sub_module_summaries.append(get_model_summary(child, size_unit=size_unit))
            sub_module_summaries[-1]['percent'] = sub_module_summaries[-1]['size'] / model_size * 100
        model_summary.update({'sub_modules': sub_module_summaries})
    return model_summary


if __name__ == '__main__':
    import json
    from dandelion.model import Alternate_2D_LSTM
    input_dim, hidden_dim, B, H, W = 8, 8, 2, 32, 32
    model = Alternate_2D_LSTM(input_dims=[input_dim],hidden_dim=hidden_dim, peephole=True, mode=2)
    model_summary = get_model_summary(model, size_unit=1)
    rpt = json.dumps(model_summary, ensure_ascii=False, indent=2)
    print(rpt)





import numpy as np
import theano
import theano.tensor as tensor
import gzip, pickle
import socket, time

def floatX(arr):
    """Converts data to a numpy array of dtype ``theano.config.floatX``.

    Parameters
    ----------
    arr : array_like
        The data to be converted.

    Returns
    -------
    numpy ndarray
        The input array in the ``floatX`` dtype configured for Theano.
        If `arr` is an ndarray of correct dtype, it is returned as is.
    """
    return np.asarray(arr, dtype=theano.config.floatX)

def shared_empty(dim=2, dtype=None):
    """Creates empty Theano shared variable.

    Shortcut to create an empty Theano shared variable with
    the specified number of dimensions.

    Parameters
    ----------
    dim : int, optional
        The number of dimensions for the empty variable, defaults to 2.
    dtype : a numpy data-type, optional
        The desired dtype for the variable. Defaults to the Theano
        ``floatX`` dtype.

    Returns
    -------
    Theano shared variable
        An empty Theano shared variable of dtype ``dtype`` with
        `dim` dimensions.
    """
    if dtype is None:
        dtype = theano.config.floatX

    shp = tuple([1] * dim)
    return theano.shared(np.zeros(shp, dtype=dtype))

def as_theano_expression(input):
    """Wrap as Theano expression.

    Wraps the given input as a Theano constant if it is not
    a valid Theano expression already. Useful to transparently
    handle numpy arrays and Python scalars, for example.

    Parameters
    ----------
    input : number, numpy array or Theano expression
        Expression to be converted to a Theano constant.

    Returns
    -------
    Theano symbolic constant
        Theano constant version of `input`.
    """
    if isinstance(input, theano.gof.Variable):
        return input
    else:
        try:
            return theano.tensor.constant(input)
        except Exception as e:
            raise TypeError("Input of type %s is not a Theano expression and "
                            "cannot be wrapped as a Theano constant (original "
                            "exception: %s)" % (type(input), e))

def collect_shared_vars(expressions):
    """Returns all shared variables the given expression(s) depend on.

    Parameters
    ----------
    expressions : Theano expression or iterable of Theano expressions
        The expressions to collect shared variables from.

    Returns
    -------
    list of Theano shared variables
        All shared variables the given expression(s) depend on, in fixed order
        (as found by a left-recursive depth-first search). If some expressions
        are shared variables themselves, they are included in the result.
    """
    # wrap single expression in list
    if isinstance(expressions, theano.Variable):
        expressions = [expressions]
    # return list of all shared variables
    return [v for v in theano.gof.graph.inputs(reversed(expressions))
            if isinstance(v, theano.compile.SharedVariable)]

def one_hot(x, m=None):
    """One-hot representation of integer vector.

    Given a vector of integers from 0 to m-1, returns a matrix
    with a one-hot representation, where each row corresponds
    to an element of x.

    Parameters
    ----------
    x : integer vector
        The integer vector to convert to a one-hot representation.
    m : int, optional
        The number of different columns for the one-hot representation. This
        needs to be strictly greater than the maximum value of `x`.
        Defaults to ``max(x) + 1``.

    Returns
    -------
    Theano tensor variable
        A Theano tensor variable of shape (``n``, `m`), where ``n`` is the
        length of `x`, with the one-hot representation of `x`.

    Notes
    -----
    If your integer vector represents target class memberships, and you wish to
    compute the cross-entropy between predictions and the target class
    memberships, then there is no need to use this function, since the function
    :func:`lasagne.objectives.categorical_crossentropy()` can compute the
    cross-entropy from the integer vector directly.

    """
    if m is None:
        m = tensor.cast(tensor.max(x) + 1, 'int32')

    return tensor.eye(m)[tensor.cast(x, 'int32')]

def batch_gather(x, index):
    """

    :param x: (B, N)
    :param index: (B,)
    :return:
    """
    B, N = x.shape
    flat_index = tensor.arange(0, B) * N + tensor.flatten(index)
    x = tensor.flatten(x)
    return x[flat_index]

def unique(l):
    """Filters duplicates of iterable.

    Create a new list from l with duplicate entries removed,
    while preserving the original order.

    Parameters
    ----------
    l : iterable
        Input iterable to filter of duplicates.

    Returns
    -------
    list
        A list of elements of `l` without duplicates and in the same order.
    """
    new_list = []
    seen = set()
    for el in l:
        if el not in seen:
            new_list.append(el)
            seen.add(el)

    return new_list

def as_tuple(x, N, t=None):
    """
    Coerce a value to a tuple of given length (and possibly given type).

    Parameters
    ----------
    x : value or iterable
    N : integer
        length of the desired tuple
    t : type, optional
        required type for all elements

    Returns
    -------
    tuple
        ``tuple(x)`` if `x` is iterable, ``(x,) * N`` otherwise.

    Raises
    ------
    TypeError
        if `type` is given and `x` or any of its elements do not match it
    ValueError
        if `x` is iterable, but does not have exactly `N` elements
    """
    try:
        X = tuple(x)
    except TypeError:
        X = (x,) * N

    if (t is not None) and not all(isinstance(v, t) for v in X):
        raise TypeError("expected a single value or an iterable "
                        "of {0}, got {1} instead".format(t.__name__, x))

    if len(X) != N:
        raise ValueError("expected a single value or an iterable "
                         "with length {0}, got {1} instead".format(N, x))

    return X

def inspect_kwargs(func):
    """
    Inspects a callable and returns a list of all optional keyword arguments.

    Parameters
    ----------
    func : callable
        The callable to inspect

    Returns
    -------
    kwargs : list of str
        Names of all arguments of `func` that have a default value, in order
    """
    # We try the Python 3.x way first, then fall back to the Python 2.x way
    try:
        from inspect import signature
    except ImportError:  # pragma: no cover
        from inspect import getargspec
        spec = getargspec(func)
        return spec.args[-len(spec.defaults):] if spec.defaults else []
    else:  # pragma: no cover
        params = signature(func).parameters
        return [p.name for p in params.values() if p.default is not p.empty]

def compute_norms(array, norm_axes=None):
    """ Compute incoming weight vector norms.

    Parameters
    ----------
    array : numpy array or Theano expression
        Weight or bias.
    norm_axes : sequence (list or tuple)
        The axes over which to compute the norm.  This overrides the
        default norm axes defined for the number of dimensions
        in `array`. When this is not specified and `array` is a 2D array,
        this is set to `(0,)`. If `array` is a 3D, 4D or 5D array, it is
        set to a tuple listing all axes but axis 0. The former default is
        useful for working with dense layers, the latter is useful for 1D,
        2D and 3D convolutional layers.
        Finally, in case `array` is a vector, `norm_axes` is set to an empty
        tuple, and this function will simply return the absolute value for
        each element. This is useful when the function is applied to all
        parameters of the network, including the bias, without distinction.
        (Optional)

    Returns
    -------
    norms : 1D array or Theano vector (1D)
        1D array or Theano vector of incoming weight/bias vector norms.

    Examples
    --------
    >>> array = np.random.randn(100, 200)
    >>> norms = compute_norms(array)
    >>> norms.shape
    (200,)

    >>> norms = compute_norms(array, norm_axes=(1,))
    >>> norms.shape
    (100,)
    """

    # Check if supported type
    if not isinstance(array, theano.Variable) and \
       not isinstance(array, np.ndarray):
        raise RuntimeError(
            "Unsupported type {}. "
            "Only theano variables and numpy arrays "
            "are supported".format(type(array))
        )

    # Compute default axes to sum over
    ndim = array.ndim
    if norm_axes is not None:
        sum_over = tuple(norm_axes)
    elif ndim == 1:          # For Biases that are in 1d (e.g. b of DenseLayer)
        sum_over = ()
    elif ndim == 2:          # DenseLayer
        sum_over = (0,)
    elif ndim in [3, 4, 5]:  # Conv{1,2,3}DLayer
        sum_over = tuple(range(1, ndim))
    else:
        raise ValueError(
            "Unsupported tensor dimensionality {}. "
            "Must specify `norm_axes`".format(array.ndim)
        )

    # Run numpy or Theano norm computation
    if isinstance(array, theano.Variable):
        # Apply theano version if it is a theano variable
        if len(sum_over) == 0:
            norms = tensor.abs_(array)   # abs if we have nothing to sum over
        else:
            norms = tensor.sqrt(tensor.sum(array**2, axis=sum_over))
    elif isinstance(array, np.ndarray):
        # Apply the numpy version if ndarray
        if len(sum_over) == 0:
            norms = abs(array)     # abs if we have nothing to sum over
        else:
            norms = np.sqrt(np.sum(array**2, axis=sum_over))

    return norms

def create_param(spec, shape=None, name=None, dim_broadcast=None):
    """
    Helper method to create Theano shared variables for layer parameters
    and to initialize them. Modified from Lasagne.utils.py

    Parameters
    ----------
    spec : scalar number, numpy array, Theano expression, or callable
        Either of the following:

        * a scalar or a numpy array with the initial parameter values
        * a Theano expression or shared variable representing the parameters
        * a function or callable that takes the desired shape of
          the parameter array as its single argument and returns
          a numpy array, a Theano expression, or a shared variable
          representing the parameters.

    shape : iterable of int (optional)
        a tuple or other iterable of integers representing the desired
        shape of the parameter array.

    dim_broadcast: tuple or list of boolean, indicating each dimension is broadcastable or not.
                   if None, then no dimension is explicitly set as broadcastable.

    name : string (optional)
        The name to give to the parameter variable. Ignored if `spec`
        is or returns a Theano expression or shared variable that
        already has a name.

    Returns
    -------
    Theano shared variable or Theano expression
        A Theano shared variable or expression representing layer parameters.
        If a scalar or a numpy array was provided, a shared variable is
        initialized to contain this array. If a shared variable or expression
        was provided, it is simply returned. If a callable was provided, it is
        called, and its output is used to initialize a shared variable.
    """
    import numbers  # to check if argument is a number
    if shape is not None:
        shape = tuple(shape)  # convert to tuple if needed
        if any(d <= 0 for d in shape):
            raise ValueError((
                "Cannot create param with a non-positive shape dimension. "
                "Tried to create param with shape=%r, name=%r") % (shape, name))

    err_prefix = "cannot initialize parameter %s: " % name
    if callable(spec):
        spec = spec(shape)
        err_prefix += "the %s returned by the provided callable"
    else:
        err_prefix += "the provided %s"

    if isinstance(spec, numbers.Number) or isinstance(spec, np.generic) \
            and spec.dtype.kind in 'biufc':
        spec = np.asarray(spec)

    if isinstance(spec, np.ndarray):
        if shape is not None and spec.shape != shape:
            raise ValueError("%s has shape %s, should be %s" %
                             (err_prefix % "numpy array", spec.shape, shape))
        if dim_broadcast is None:
            spec = theano.shared(spec)
        else:
            spec = theano.shared(spec, broadcastable=dim_broadcast)
        # if shape is None:
            # spec = theano.shared(spec)
        # else:
            # bcast = tuple(s == 1 for s in shape)
            # spec = theano.shared(spec, broadcastable=bcast)

    # if dim_broadcast is not None:
        # spec = tensor.patternbroadcast(spec, dim_broadcast)

    if isinstance(spec, theano.Variable):
        # We cannot check the shape here, Theano expressions (even shared
        # variables) do not have a fixed compile-time shape. We can check the
        # dimensionality though.
        if shape is not None and spec.ndim != len(shape):
            raise ValueError("%s has %d dimensions, should be %d" %
                             (err_prefix % "Theano variable", spec.ndim,
                              len(shape)))
        # We only assign a name if the user hasn't done so already.
        if not spec.name:
            spec.name = name
        return spec

    else:
        if "callable" in err_prefix:
            raise TypeError("%s is not a numpy array or a Theano expression" %
                            (err_prefix % "value"))
        else:
            raise TypeError("%s is not a numpy array, a Theano expression, "
                            "or a callable" % (err_prefix % "spec"))

def unroll_scan(fn, sequences, outputs_info, non_sequences, n_steps,
                go_backwards=False):
        """
        Helper function to unroll for loops. Can be used to unroll theano.scan.
        The parameter names are identical to theano.scan, please refer to here
        for more information.

        Note that this function does not support the truncate_gradient
        setting from theano.scan.

        Parameters
        ----------

        fn : function
            Function that defines calculations at each step.

        sequences : TensorVariable or list of TensorVariables
            List of TensorVariable with sequence data. The function iterates
            over the first dimension of each TensorVariable.

        outputs_info : list of TensorVariables
            List of tensors specifying the initial values for each recurrent
            value.

        non_sequences: list of TensorVariables
            List of theano.shared variables that are used in the step function.

        n_steps: int
            Number of steps to unroll.

        go_backwards: bool
            If true the recursion starts at sequences[-1] and iterates
            backwards.

        Returns
        -------
        List of TensorVariables. Each element in the list gives the recurrent
        values at each time step.

        """
        if not isinstance(sequences, (list, tuple)):
            sequences = [sequences]

        # When backwards reverse the recursion direction
        counter = range(n_steps)
        if go_backwards:
            counter = counter[::-1]

        output = []
        prev_vals = outputs_info
        for i in counter:
            step_input = [s[i] for s in sequences] + prev_vals + non_sequences
            out_ = fn(*step_input)
            # The returned values from step can be either a TensorVariable,
            # a list, or a tuple.  Below, we force it to always be a list.
            if isinstance(out_, tensor.TensorVariable):
                out_ = [out_]
            if isinstance(out_, tuple):
                out_ = list(out_)
            output.append(out_)

            prev_vals = output[-1]

        # iterate over each scan output and convert it to same format as scan:
        # [[output11, output12,...output1n],
        # [output21, output22,...output2n],...]
        output_scan = []
        for i in range(len(output[0])):
            l = map(lambda x: x[i], output)
            output_scan.append(tensor.stack(*l))

        return output_scan

class chunked_byte_writer(object):
    """
    This class is used for by-passing the bug in gzip/zlib library: when data length exceeds unsigned int limit, gzip/zlib will break
    file: a file object
    """
    def __init__(self, file, chunksize=4294967295):
        self.file = file
        self.chunksize = chunksize
    def write(self, data):
        for i in range(0, len(data), self.chunksize):
            self.file.write(data[i:i+self.chunksize])

class gpickle(object):
    """
    A pickle class with gzip enabled
    """
    @staticmethod
    def dump(data, filename, compresslevel=9):
        with gzip.open(filename, mode='wb', compresslevel=compresslevel) as f:
            pickle.dump(data, chunked_byte_writer(f))
            f.close()
    @staticmethod
    def load(filename):
        """
        The chunked read mechanism here is for by-passing the bug in gzip/zlib library: when
        data length exceeds unsigned int limit, gzip/zlib will break
        :param filename:
        :return:
        """
        buf = b''
        chunk = b'NULL'
        with gzip.open(filename, mode='rb') as f:
            while len(chunk) > 0:
                chunk = f.read(429496729)
                buf += chunk
        data = pickle.loads(buf)
        return data

    @staticmethod
    def loads(buf):
        return pickle.loads(buf)

    @staticmethod
    def dumps(data):
        return pickle.dumps(data)

def get_weight_by_name(module_weights, name):
    """
    Retrieve parameter weight values from result of Module.get_weights()
    :param module_weights: all weights of a module, returned by Module.get_weights()
    :param name:
    :return:
    """
    for w, w_name in module_weights:
        if name == w_name:
            return w
    return None

def theano_safe_run(fn, input_list):
    """
    Help catch theano memory exceptions during running theano functions.
    :param fn:
    :param input_list:
    :return: (status, result), status > 0 means exception catched.
    """
    try:
        result = fn(*input_list)
        status = 0
        return status, result
    except MemoryError:
        print('Memory error catched')
        status = 1
        return status, None
    except RuntimeError as e:
        print('RuntimeError encountered')
        if e.args[0].startswith('CudaNdarray_ZEROS: allocation failed.'):
            print('Memory error catched')
            status = 2
            return status, None
        elif str(e).startswith('gpudata_alloc: cuMemAlloc: CUDA_ERROR_OUT_OF_MEMORY: out of memory'):
            print('Memory error catched')
            status = 3
            return status, None
        else:
            raise e
    except Exception as e:
        if e.args[0].startswith("b'cuMemAlloc: CUDA_ERROR_OUT_OF_MEMORY: out of memory'"):
            print('New backend memory error catched')
            status = 4
            return status, None
        else:
            raise e

class Finite_Memory_Array(object):

    def __init__(self, array=None, shape=None, dtype=np.double):
        if array is not None:
            self.array = array
        elif shape is not None:
            self.array = np.zeros(shape=shape, dtype=dtype)
        else:
            raise ValueError('bad initialization para for Finite_Memory_Array')
        self.shape = self.array.shape
        self.curpos = 0
        self.first_round = True

    def update(self, value):
        self.array[:, self.curpos] = value
        self.curpos += 1
        if self.curpos >= self.shape[1]:
            self.curpos = 0
            self.first_round = False

    def clear(self):
        self.array[:] = 0
        self.curpos = 0
        self.first_round = True

    def get_current_position(self):
        return self.curpos

    @property
    def content(self):
        if self.first_round is True:
            return self.array[:, :self.curpos]
        else:
            return self.array

def pad_sequence_into_array(Xs, maxlen=None, truncating='post', padding='post', value=0.):
    """
    Padding sequence (list of numpy arrays) into an numpy array
    :param Xs: list of numpy arrays. The arrays must have the same shape except the first dimension.
    :param maxlen: the allowed maximum of the first dimension of Xs's arrays. Any array longer than maxlen is truncated to maxlen
    :param truncating: = 'pre'/'post', indicating whether the truncation happens at either the beginning or the end of the array (default)
    :param padding: = 'pre'/'post',indicating whether the padding happens at either the beginning or the end of the array (default)
    :param value: scalar, the padding value, default = 0.0
    :return: Xout, the padded sequence (now an augmented array with shape (Narrays, N1stdim, N2nddim, ...)
    :return: mask, the corresponding mask, binary array, with shape (Narray, N1stdim)
    """
    Nsamples = len(Xs)
    if maxlen is None:
        lengths = [s.shape[0] for s in Xs]    # 'sequences' must be list, 's' must be numpy array, len(s) return the first dimension of s
        maxlen = np.max(lengths)

    Xout = np.ones(shape = [Nsamples, maxlen] + list(Xs[0].shape[1:]), dtype=Xs[0].dtype) * np.asarray(value, dtype=Xs[0].dtype)
    Mask = np.zeros(shape = [Nsamples, maxlen], dtype=Xout.dtype)
    for i in range(Nsamples):
        x = Xs[i]
        if truncating == 'pre':
            trunc = x[-maxlen:]
        elif truncating == 'post':
            trunc = x[:maxlen]
        else:
            raise ValueError("Truncating type '%s' not understood" % truncating)
        if padding == 'post':
            Xout[i, :len(trunc)] = trunc
            Mask[i, :len(trunc)] = 1
        elif padding == 'pre':
            Xout[i, -len(trunc):] = trunc
            Mask[i, -len(trunc):] = 1
        else:
            raise ValueError("Padding type '%s' not understood" % padding)
    return Xout, Mask

def get_local_ip():
    """
    Get local host IP address
    :return:
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 0))                                # connecting to a UDP address doesn't send packets
    local_ip_address = s.getsockname()[0]
    return local_ip_address

def get_time_stamp():
    """
    Create a formatted string time stamp
    :return:
    """
    return time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime())


if __name__ == '__main__':
    INFO = ['This is a collection of auxiliary functions for DL.\n',
            'Author: David Leon\n',
            'All rights reserved\n']
    print(*INFO)

    # x1 = np.array([1,2,3])
    # x2 = np.array([1,2,3,4])
    # x3 = np.array([5,4,3,2,1])
    x1, x2, x3 = np.random.rand(2, 4), np.random.rand(3, 4), np.random.rand(5,4)

    Xout, mask = pad_sequence_into_array([x1,x2,x3])
    print(Xout)
    print(mask)
    print(Xout.shape)
    print(mask.shape)


# coding:utf-8
# Test for softmax/log_softmax for ndim > 2.
# This test is added because occasionally Theano compiled function would raise shape mismatch error for reshap op. inside softmax/log_softmax when ndim >2.
# Created   :   7, 23, 2018
# Revised   :   7, 23, 2018
# All rights reserved
#------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'
import os, numpy as np
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
from theano import tensor
from dandelion.module import *
from dandelion.activation import softmax, log_softmax
from dandelion.objective import categorical_crossentropy, categorical_crossentropy_log
from dandelion.util import one_hot

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

# ndim = 2
def test_case_0():
    x = tensor.fmatrix()  # (B, N)
    y = softmax(x)  # (B, N)
    f = theano.function([x] , y)

    for i in range(10):
        B = 1
        N = 5
        x = np.random.rand(B, N).astype(np.float32)
        y = f(x)
        print(y.shape)

# ndim = 3
def test_case_1():
    x = tensor.ftensor3()  # (B, T, N)
    y = softmax(x)  # (B, T, N)
    f = theano.function([x] , y)

    for i in range(20):
        B = np.random.randint(1, 10)
        T = np.random.randint(1, 10)
        N = np.random.randint(2, 8)
        x = np.random.rand(B, T, N).astype(np.float32)
        y = f(x)
        print(y.shape)

# ndim = 4
def test_case_2():
    x = tensor.ftensor4()  # (B, H, W, N)
    y = softmax(x)  # (B, H, W, N)
    f = theano.function([x], y)

    for i in range(20):
        B = np.random.randint(1, 10)
        H = np.random.randint(1, 100)
        W = np.random.randint(1, 100)
        N = np.random.randint(2, 8)
        x = np.random.rand(B, H, W, N).astype(np.float32)
        y = f(x)
        print(y.shape)

if __name__ == '__main__':

    test_case_0()

    test_case_1()

    test_case_2()

    print('Test passed')






# coding:utf-8
'''
Unitest for AlternateLSTM2D model
Created   :  11, 13, 2018
Revised   :  11, 13, 2018
All rights reserved
'''
# ------------------------------------------------------------------------------------------------
__author__ = 'dawei.leng'

import os
os.environ['THEANO_FLAGS'] = "floatX=float32, mode=FAST_RUN, warn_float64='raise'"

import theano
import theano.tensor as tensor
from dandelion.model import Alternate_2D_LSTM
from dandelion.objective import *
from dandelion.update import *
import numpy as np, time

import dandelion
dandelion_path = os.path.split(dandelion.__file__)[0]
print('dandelion path = %s\n' % dandelion_path)

def test_case_0():
    # input_dim, hidden_dim, B, H, W = 5, 4, 2, 5, 7
    input_dim, hidden_dim, B, H, W = 8, 8, 2, 32, 32
    input = tensor.ftensor4()
    gt    = tensor.ftensor4()
    model = Alternate_2D_LSTM(input_dims=[input_dim],hidden_dim=hidden_dim, peephole=True, mode=2)
    output = model.forward(input, backward=False)
    loss = aggregate(squared_error(output, gt))
    params = model.collect_params()
    updates = sgd(loss, params, 1e-4)
    updates.update(model.collect_self_updates())
    print('compiling function ...')
    f = theano.function([input, gt], [output, loss], updates=updates, no_default_updates=False)

    print('run function ...')
    X = np.random.rand(H, W, B, input_dim).astype('float32')
    GT = np.random.rand(H, W, B, hidden_dim).astype('float32')
    time0 = time.time()
    Y, loss = f(X, GT)
    time_used = time.time() - time0
    print('time_used = ', time_used)
    print('Y=', Y)
    print('Y.shape=', Y.shape)
    print('loss=', loss)

if __name__ == '__main__':

    test_case_0()

    print('Test passed')